

#if defined(__SAM3X8E__)
	#include <include/pio.h>
    #define PROGMEM
    #define pgm_read_byte(addr) (*(const unsigned char *)(addr))
    #define pgm_read_word(addr) (*(const unsigned short *)(addr))
#endif
#ifdef __AVR__
	#include <avr/pgmspace.h>
#endif
#include "pins_arduino.h"
#include "wiring_private.h"
#include "ILI9488.h"
#include "pin_magic.h"

#define TFTWIDTH   320    
#define TFTHEIGHT  480


#define TFTREGISTER_SOFTRESET          0x01
#define TFTREGISTER_SLEEPIN            0x10
#define TFTREGISTER_SLEEPOUT           0x11
#define TFTREGISTER_NORMALDISP         0x13
#define TFTREGISTER_INVERTOFF          0x20
#define TFTREGISTER_INVERTON           0x21
#define TFTREGISTER_GAMMASET           0x26
#define TFTREGISTER_DISPLAYOFF         0x28
#define TFTREGISTER_DISPLAYON          0x29
#define TFTREGISTER_COLADDRSET         0x2A
#define TFTREGISTER_PAGEADDRSET        0x2B
#define TFTREGISTER_MEMORYWRITE        0x2C
#define TFTREGISTER_PIXELFORMAT        0x3A
#define TFTREGISTER_FRAMECONTROL       0xB1
#define TFTREGISTER_DISPLAYFUNC        0xB6
#define TFTREGISTER_ENTRYMODE          0xB7
#define TFTREGISTER_POWERCONTROL1      0xC0
#define TFTREGISTER_POWERCONTROL2      0xC1
#define TFTREGISTER_VCOMCONTROL1      0xC5
#define TFTREGISTER_VCOMCONTROL2      0xC7
#define TFTREGISTER_MEMCONTROL      0x36

#define TFTREGISTER_MEMCONTROL_MY  0x80
#define TFTREGISTER_MEMCONTROL_MX  0x40
#define TFTREGISTER_MEMCONTROL_MV  0x20
#define TFTREGISTER_MEMCONTROL_ML  0x10
#define TFTREGISTER_MEMCONTROL_RGB 0x00
#define TFTREGISTER_MEMCONTROL_BGR 0x08
#define TFTREGISTER_MEMCONTROL_MH  0x04


// Constructor for breakout board (configurable LCD control lines).
// Can still use this w/shield, but parameters are ignored.
ILI9488::ILI9488(
  uint8_t cs, uint8_t cd, uint8_t wr, uint8_t rd, uint8_t reset) :
  Adafruit_GFX(TFTWIDTH, TFTHEIGHT) {

#ifndef USE_ADAFRUIT_SHIELD_PINOUT
  // Convert pin numbers to registers and bitmasks
  _reset     = reset;
  #ifdef __AVR__
    csPort     = portOutputRegister(digitalPinToPort(cs));
    cdPort     = portOutputRegister(digitalPinToPort(cd));
    wrPort     = portOutputRegister(digitalPinToPort(wr));
    rdPort     = portOutputRegister(digitalPinToPort(rd));
  #endif
  #if defined(__SAM3X8E__)
    csPort     = digitalPinToPort(cs);
    cdPort     = digitalPinToPort(cd);
    wrPort     = digitalPinToPort(wr);
    rdPort     = digitalPinToPort(rd);
  #endif
  csPinSet   = digitalPinToBitMask(cs);
  cdPinSet   = digitalPinToBitMask(cd);
  wrPinSet   = digitalPinToBitMask(wr);
  rdPinSet   = digitalPinToBitMask(rd);
  csPinUnset = ~csPinSet;
  cdPinUnset = ~cdPinSet;
  wrPinUnset = ~wrPinSet;
  rdPinUnset = ~rdPinSet;
  #ifdef __AVR__
    *csPort   |=  csPinSet; // Set all control bits to HIGH (idle)
    *cdPort   |=  cdPinSet; // Signals are ACTIVE LOW
    *wrPort   |=  wrPinSet;
    *rdPort   |=  rdPinSet;
  #endif
  #if defined(__SAM3X8E__)
    csPort->PIO_SODR  |=  csPinSet; // Set all control bits to HIGH (idle)
    cdPort->PIO_SODR  |=  cdPinSet; // Signals are ACTIVE LOW
    wrPort->PIO_SODR  |=  wrPinSet;
    rdPort->PIO_SODR  |=  rdPinSet;
  #endif
  pinMode(cs, OUTPUT);    // Enable outputs
  pinMode(cd, OUTPUT);
  pinMode(wr, OUTPUT);
  pinMode(rd, OUTPUT);
  if(reset) {
    digitalWrite(reset, HIGH);
    pinMode(reset, OUTPUT);
  }
#endif

  init();
}

// Constructor for shield (fixed LCD control lines)
ILI9488::ILI9488(void) : Adafruit_GFX(TFTWIDTH, TFTHEIGHT) {
  init();
}

// Initialization common to both shield & breakout configs
void ILI9488::init(void) {

#ifdef USE_ADAFRUIT_SHIELD_PINOUT
  CS_IDLE; // Set all control bits to idle state
  WR_IDLE;
  RD_IDLE;
  CD_DATA;
  digitalWrite(5, HIGH); // Reset line
  pinMode(A3, OUTPUT);   // Enable outputs
  pinMode(A2, OUTPUT);
  pinMode(A1, OUTPUT);
  pinMode(A0, OUTPUT);
  pinMode( 5, OUTPUT);
#endif

  setWriteDir(); // Set up LCD data port(s) for WRITE operations

  rotation  = 0;
  cursor_y  = cursor_x = 0;
  textsize  = 1;
  textcolor = 0xFFFF;
  _width    = TFTWIDTH;
  _height   = TFTHEIGHT;
}

// Initialization command tables for different LCD controllers
#define TFTLCD_DELAY 0xFF

static const uint8_t ILI9488_regValues[] PROGMEM = {

	0x11, 0,                   // sleep OUT
	TFTLCD_DELAY, 150,
	0x38, 0,                   // idle mode OFF
	0x13, 0,                   // normal display mode ON
	0x55, 1, 0x02,             // High enhancment of CE ratio   Write Content Adaptive Brightness Control
	0x34, 0,                   
	0xB1, 2, 0xA0, 0x10,
	0xB4, 1, 0x10,
	0xB5, 2, 0x00, 0x12,
	0xB6, 3, 0x00, 0x21, 0x3B, 
	0xC0, 2, 0x1F, 0x01,
	0xC1, 2, 0x00, 0x10,
	0xC2, 1, 0x00,              //0X44 lower better
	0xC4, 1, 0x00,              //0X10
	0xC5, 3, 0x00, 0x80, 0x80,
	0x3A, 1, 0x55,
	0x2A, 4, 0x00, 0x00, 0x01, 0x3F, //320
	0x2B, 4, 0x00, 0x00, 0x01, 0xDF, //480
	0x29, 0,                         //Display ON
	TFTLCD_DELAY, 30,
	0x2C, 0,                         //Memory Write
	TFTLCD_DELAY, 30,  

};

void ILI9488::begin() {

  uint16_t id = 0x9488;
  uint8_t i = 0;

  reset();

  delay(200);

	CS_ACTIVE;
	while(i < sizeof(ILI9488_regValues)) {
		uint8_t r = pgm_read_byte(&ILI9488_regValues[i++]);
		uint8_t len = pgm_read_byte(&ILI9488_regValues[i++]);
		if(r == TFTLCD_DELAY) {
			delay(len);
		} 
    else {
			//Serial.print("Register $"); Serial.print(r, HEX);  Serial.print(" -> "); 

			CS_ACTIVE;
			CD_COMMAND;
			write8(r);
			CD_DATA;
			for (uint8_t d=0; d<len; d++) {
				uint8_t x = pgm_read_byte(&ILI9488_regValues[i++]);
				write8(x);
			  //Serial.print(x, HEX); Serial.print(" ");
			}
			CS_IDLE;
      //Serial.println();
		}
	}
	
	setRotation (0);
  return;	

}

void ILI9488::reset(void) {

  CS_IDLE;
//  CD_DATA;
  WR_IDLE;
  RD_IDLE;

#ifdef USE_ADAFRUIT_SHIELD_PINOUT
  digitalWrite(5, LOW);
  delay(2);
  digitalWrite(5, HIGH);
#else
  if(_reset) {
    digitalWrite(_reset, LOW);
    delay(2);
    digitalWrite(_reset, HIGH);
  }
#endif

  // Data transfer sync
  CS_ACTIVE;
  CD_COMMAND;
  write8(0x00);
  for(uint8_t i=0; i<3; i++) WR_STROBE; // Three extra 0x00s
  CS_IDLE;
}

void ILI9488::setAddrWindow(int x1, int y1, int x2, int y2) {
  CS_ACTIVE;
  uint32_t t;

  t = x1;
  t <<= 16;
  t |= x2;
  writeRegister32(TFTREGISTER_COLADDRSET, t);  // HX8357D uses same registers! 
  t = y1;
  t <<= 16;
  t |= y2;
  writeRegister32(TFTREGISTER_PAGEADDRSET, t); // HX8357D uses same registers!
  
  CS_IDLE;
}



void ILI9488::flood(uint16_t color, uint32_t len) {
  uint16_t blocks;
  uint8_t  i, hi = color >> 8,
              lo = color;

  CS_ACTIVE;
  CD_COMMAND;
  write8(0x2C);

  // Write first pixel normally, decrement counter by 1
  CD_DATA;
  write8(hi);
  write8(lo);
  len--;

  blocks = (uint16_t)(len / 64); // 64 pixels/block
  if(hi == lo) {
    // High and low bytes are identical.  Leave prior data
    // on the port(s) and just toggle the write strobe.
    while(blocks--) {
      i = 16; // 64 pixels/block / 4 pixels/pass
      do {
        WR_STROBE; WR_STROBE; WR_STROBE; WR_STROBE; // 2 bytes/pixel
        WR_STROBE; WR_STROBE; WR_STROBE; WR_STROBE; // x 4 pixels
      } while(--i);
    }
    // Fill any remaining pixels (1 to 64)
    for(i = (uint8_t)len & 63; i--; ) {
      WR_STROBE;
      WR_STROBE;
    }
  } else {
    while(blocks--) {
      i = 16; // 64 pixels/block / 4 pixels/pass
      do {
        write8(hi); write8(lo); write8(hi); write8(lo);
        write8(hi); write8(lo); write8(hi); write8(lo);
      } while(--i);
    }
    for(i = (uint8_t)len & 63; i--; ) {
      write8(hi);
      write8(lo);
    }
  }
  CS_IDLE;
}

void ILI9488::drawFastHLine(int16_t x, int16_t y, int16_t length,
  uint16_t color)
{
  int16_t x2;

  // Initial off-screen clipping
  if((length <= 0     ) ||
     (y      <  0     ) || ( y                  >= _height) ||
     (x      >= _width) || ((x2 = (x+length-1)) <  0      )) return;

  if(x < 0) {        // Clip left
    length += x;
    x       = 0;
  }
  if(x2 >= _width) { // Clip right
    x2      = _width - 1;
    length  = x2 - x + 1;
  }

  setAddrWindow(x, y, x2, y);
  flood(color, length);

}

void ILI9488::drawFastVLine(int16_t x, int16_t y, int16_t length,
  uint16_t color)
{
  int16_t y2;

  // Initial off-screen clipping
  if((length <= 0      ) ||
     (x      <  0      ) || ( x                  >= _width) ||
     (y      >= _height) || ((y2 = (y+length-1)) <  0     )) return;
  if(y < 0) {         // Clip top
    length += y;
    y       = 0;
  }
  if(y2 >= _height) { // Clip bottom
    y2      = _height - 1;
    length  = y2 - y + 1;
  }

  setAddrWindow(x, y, x, y2);
  flood(color, length);

}

void ILI9488::fillRect(int16_t x1, int16_t y1, int16_t w, int16_t h, 
  uint16_t fillcolor) {
  int16_t  x2, y2;

  // Initial off-screen clipping
  if( (w            <= 0     ) ||  (h             <= 0      ) ||
      (x1           >= _width) ||  (y1            >= _height) ||
     ((x2 = x1+w-1) <  0     ) || ((y2  = y1+h-1) <  0      )) return;
  if(x1 < 0) { // Clip left
    w += x1;
    x1 = 0;
  }
  if(y1 < 0) { // Clip top
    h += y1;
    y1 = 0;
  }
  if(x2 >= _width) { // Clip right
    x2 = _width - 1;
    w  = x2 - x1 + 1;
  }
  if(y2 >= _height) { // Clip bottom
    y2 = _height - 1;
    h  = y2 - y1 + 1;
  }

  setAddrWindow(x1, y1, x2, y2);
  flood(fillcolor, (uint32_t)w * (uint32_t)h);

}

void ILI9488::fillScreen(uint16_t color) {  
  setAddrWindow(0, 0, _width - 1, _height - 1);
  flood(color, (long)TFTWIDTH * (long)TFTHEIGHT);
}

void ILI9488::drawPixel(int16_t x, int16_t y, uint16_t color) {

  // Clip
  if((x < 0) || (y < 0) || (x >= _width) || (y >= _height)) return;

  CS_ACTIVE;
  setAddrWindow (x, y, x, y);
  CS_ACTIVE;
  CD_COMMAND; 
  write8(0x2C);
  CD_DATA; 
  write8(color >> 8); 
  write8(color);
  CS_IDLE;
}

void ILI9488::pushColors(uint16_t *data, uint8_t len, boolean first) {
  uint16_t color;
  uint8_t  hi, lo;
  CS_ACTIVE;
  if(first == true) { // Issue GRAM write command only on first call
    CD_COMMAND;
    write8(0x2C);
    CD_DATA;
    while(len--) {
      color = *data++;
      hi    = color >> 8; // Don't simplify or merge these
      lo    = color;      // lines, there's macro shenanigans
      write8(hi);         // going on.
      write8(lo);
    }
    CS_IDLE;
  }
}

void ILI9488::setRotation(uint8_t x) {

  // Call parent rotation func first -- sets up rotation flags, etc.
  Adafruit_GFX::setRotation(x);
  // Then perform hardware-specific rotation operations...

  CS_ACTIVE;
  
  uint8_t t; 
  switch (rotation) {
    case 2:
      t = TFTREGISTER_MEMCONTROL_MX | TFTREGISTER_MEMCONTROL_MY | TFTREGISTER_MEMCONTROL_BGR;
      break;
    case 3:
      t = TFTREGISTER_MEMCONTROL_MV | TFTREGISTER_MEMCONTROL_MY | TFTREGISTER_MEMCONTROL_BGR;
      break;
    case 0:
      t = TFTREGISTER_MEMCONTROL_BGR;
      break;
    case 1:
      t = TFTREGISTER_MEMCONTROL_MX | TFTREGISTER_MEMCONTROL_MV | TFTREGISTER_MEMCONTROL_BGR;
      break;
  }
  writeRegister8(TFTREGISTER_MEMCONTROL, t); // MEMCONTROL

  setAddrWindow(0, 0, _width - 1, _height - 1); // CS_IDLE happens here

  CS_IDLE;
}

#ifdef read8isFunctionalized
  #define read8(x) x=read8fn()
#endif

uint16_t ILI9488::readPixel(int16_t x, int16_t y) {

  if((x < 0) || (y < 0) || (x >= _width) || (y >= _height)) return 0;
  return 0;
}

// Ditto with the read/write port directions, as above.
uint16_t ILI9488::readID(void) {

  uint16_t id = readReg(0xD3);
  return id;
    
}

uint32_t ILI9488::readReg(uint8_t r) {
  uint32_t id;
  uint8_t x;

  // try reading register #4
  CS_ACTIVE;
  CD_COMMAND;
  write8(r);
  setReadDir();  // Set up LCD data port(s) for READ operations
  CD_DATA;
  delayMicroseconds(50);

  read8(x);
  id = x;          // Do not merge or otherwise simplify
  id <<= 8;              // these lines.  It's an unfortunate
  read8(x);
  id  |= x;        // shenanigans that are going on.
  id <<= 8;              // these lines.  It's an unfortunate
  read8(x);
  id  |= x;        // shenanigans that are going on.
  id <<= 8;              // these lines.  It's an unfortunate
  read8(x);
  id  |= x;        // shenanigans that are going on.
  if (r == 0xEF){
    id <<= 8;         // these lines.  It's an unfortunate
	read8(x);
	id  |= x;	  
  }
  CS_IDLE;
  setWriteDir();  // Restore LCD data port(s) to WRITE configuration
  delay(150); //stabilization time
  return id;
}

// Pass 8-bit (each) R,G,B, get back 16-bit packed color
uint16_t ILI9488::color565(uint8_t r, uint8_t g, uint8_t b) {
  return ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3);
}

// For I/O macros that were left undefined, declare function
// versions that reference the inline macros just once:

#ifndef write8
void ILI9488::write8(uint8_t value) {
  write8inline(value);
}
#endif

#ifdef read8isFunctionalized
uint8_t ILI9488::read8fn(void) {
  uint8_t result;
  read8inline(result);
  return result;
}
#endif

#ifndef setWriteDir
void ILI9488::setWriteDir(void) {
  setWriteDirInline();
}
#endif

#ifndef setReadDir
void ILI9488::setReadDir(void) {
  setReadDirInline();
}
#endif

#ifndef writeRegister8
void ILI9488::writeRegister8(uint8_t a, uint8_t d) {
  writeRegister8inline(a, d);
}
#endif

#ifndef writeRegister16

void ILI9488::writeRegister16(uint16_t a, uint16_t d) {
  writeRegister16inline(a, d);
}
#endif

#ifndef writeRegisterPair
void ILI9488::writeRegisterPair(uint8_t aH, uint8_t aL, uint16_t d) {
  writeRegisterPairInline(aH, aL, d);
}
#endif


void ILI9488::writeRegister24(uint8_t r, uint32_t d) {
  CS_ACTIVE;
  CD_COMMAND;
  write8(r);
  CD_DATA;
  delayMicroseconds(10);
  write8(d >> 16);
  delayMicroseconds(10);
  write8(d >> 8);
  delayMicroseconds(10);
  write8(d);
  CS_IDLE;

}


void ILI9488::writeRegister32(uint8_t r, uint32_t d) {
  CS_ACTIVE;
  CD_COMMAND;
  write8(r);
  CD_DATA;
  delayMicroseconds(10);
  write8(d >> 24);
  delayMicroseconds(10);
  write8(d >> 16);
  delayMicroseconds(10);
  write8(d >> 8);
  delayMicroseconds(10);
  write8(d);
  CS_IDLE;

}
